#ifndef PDU_H
#define PDU_H


#include <sys/types.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/wait.h>

#include <signal.h>
#include <list.h>

#define INTSIZE   sizeof(int)
#define DATASIZE  1024               /* maximum length          */
#define MAXNODES  20


/* The following are the types of messages communicated between 
 * physical layer & datalink layer, and client & server 
 */
enum datatypes { DATA, ERR };
enum boolean { NO, YES };


/* ****************************************************** */
/* APPLICATION LAYER DECLARATIONS                         */
/* ****************************************************** */
typedef struct {
     int  snode;                    /* source node address      */
     int  dnode;                    /* destination node address */
     char data[DATASIZE];           /* message                  */
} A_PDU_TYPE;
#define A_PDU_SIZE sizeof(A_PDU_TYPE)

/* ****************************************************** */
/* NETWORK LAYER DECLARATIONS                             */
/* ****************************************************** */
#define MAX_IP_PAYLOAD  2048    /* artificial: to avoid impl. var size pkts */
typedef int IPaddr;

typedef struct ip {
     u_char 	ip_verlen;   /* IP version and header length: (ignore) */
     u_char 	ip_tos;      /* type of service:  (ignore) */
     short 	ip_len;      /* total packet length (in octets): USED */
     short 	ip_id;       /* datagram id  : USED */
     short 	ip_fragoff;  /* fragment offset (in 8 octets) + flags: USED  */
     u_char 	ip_ttl;      /* time to live in gateway hops: (ignore)  */
     u_char 	ip_proto;    /* IP protocol: (ignore) */
     short 	ip_chksum;   /* header checksum: (ignore) */
     IPaddr     ip_src;      /* IP address of source: USED */
     IPaddr     ip_dst;      /* IP address of destination: USED */
     u_char     ip_data[MAX_IP_PAYLOAD];    /* variable length data: USED */
} N_PDU_TYPE;
#define N_PDU_SIZE sizeof(N_PDU_TYPE)

#define IP_MF 0x2000           /* more fragments bit: 8192  */
#define IP_DF 0x4000           /* don't fragment bit: (ignore) */ 
#define IP_FRAGOFF 0x1fff      /* fragment offset mask: gives you the 13 bits */

#define IPMHLEN 20   /* min IP header length (in bytes) */

/* ------- Reassembly structures ----------- */
#define IP_FQSIZE	100  /* Max number of frag queues */
#define IP_MAXNF	100  /* Max number of frags/datagram (ignore) */
#define IP_FTTL		6    /* Max time between 2 frags (5 ms)
			        or 6 ms upto granularity  */
#define FTTL_DECR	1    /* 1 ms decrement of FTTL  (ignore) */
#define FTTL_DECR_USECS	1000 /* = 1000 usecs decrement of FTTL (ignore) */

/* ipf_state flags */
#define IPFF_VALID 	1   /* contents are valid */
#define IPFF_BOGUS 	2   /* drop frags that match */
#define IPFF_FREE 	3   /* this queue is free to be allocated  */

/* Ordered list implementation (ignore) */
typedef queuet OrderedList_t;
typedef q_elt OrderedListElement_t;

/* Fragment queue and associated state variables */
/* The destination network layer will have a table of these structures */
struct ipfq {
     char 	        ipf_state; 	/* VALID, FREE or BOGUS  */
     IPaddr 		ipf_src; 	/* IP address of the source */
     short 		ipf_id; 	/* datagram id */
     int 		ipf_ttl; 	/* countdown to disposal */
     int		nfrags; 	/* tmp variable for debugging */
     OrderedList_t     *ipf_q; 		/* ordered list of fragments   */
};



/* ****************************************************** */
/* DATA LINK LAYER DECLARATIONS                           */
/* ****************************************************** */
/* data unit between datalink layer and physical layer */
/* ---------- D_PDU Definition (Used in Go Back N Labs) ----- */
#define D_INFO 0
#define D_ACK  1
#define D_NAK  2

typedef struct {
     int               curr_node;   /* address of this node */
     int               next_node;   /* address of next node */

     /* ----------- Begin: New fields (Used in Go Back N Labs)  -------- */
     int               type;        /* One of D_INFO, D_ACK, D_NAK   */
     int               seq_number;  /* Sequence number of this D_PDU */
     int               ack_number;  /* Sequence number of ACK/NAK    */
     /* ----------- End: New fields (Used in Go Back N Labs) -------- */


     N_PDU_TYPE	       n_pdu;
     enum boolean      error;       /* YES if the packet is corrupted;
				       otherwise NO.            */
} D_PDU_TYPE;
#define D_PDU_SIZE sizeof(D_PDU_TYPE)

/* ****************************************************** */
/* PHYSICAL LAYER DECLARATIONS                            */
/* ****************************************************** */
/* data unit between physical layers */
typedef struct {
     int               type;        /* ERR, DATA                */
     D_PDU_TYPE        d_pdu;
} PH_PDU_TYPE;
#define PH_PDU_SIZE sizeof(PH_PDU_TYPE)

/* ****************************************************** */
/* PDU DECLARATIONS                                       */
/* ****************************************************** */

/* ------ Values for the type field in PDU_TYPE ---- */
#define TYPE_IS_A_PDU 0
#define TYPE_IS_N_PDU 1
#define TYPE_IS_D_PDU 2
#define TYPE_IS_PH_PDU 3

typedef struct {
     union {
	  A_PDU_TYPE a_pdu;
	  N_PDU_TYPE n_pdu;
	  D_PDU_TYPE d_pdu;
	  PH_PDU_TYPE ph_pdu;
     } u; 

     int type;    /* One of TYPE_IS_A_PDU, TYPE_IS_N_PDU,
		     TYPE_IS_D_PDU, TYPE_IS_PH_PDU */

} PDU_TYPE;

#define PDU_SIZE sizeof(PDU_TYPE)

int pdu_init(); /* Initialize the pool of packets. Returns 0 for failure. */
PDU_TYPE *pdu_alloc(); /* Allocates a packet and returns a pointer
				thereto.  The packet is guaranteed to contain
				garbage. */
void pdu_free(PDU_TYPE *pdu);       /* Passed a packet pointer, frees it. */




/* End Additions */
/**************************************************************************/

#endif   /* PDU_H   */























